'''
poetry.py, by Danielle Sucher  
https://github.com/DanielleSucher/Nantucket/
[Note: this file and its imports are being used as a library and are not part of my work for the CS224U project -- Luke]
'''

from __future__ import division
import re
import urllib2
#from curses.ascii import isdigit #This is unix-only and can be replaced with the string.isdigit() method
from nltk.corpus import cmudict
d = cmudict.dict()
from nltk.corpus.util import LazyCorpusLoader
from nltk.corpus.reader import *
suffdict = LazyCorpusLoader(
    'cmusuffdict', CMUDictCorpusReader, ['cmusuffdict'])
suffdict = suffdict.dict()


def try_syllable(syl):
    ''' helper function for phonemes()
    Tests if syl is in suffdict. If not, removes the first letter
    and then the first two letters, checking each time
    '''
    if syl in suffdict:
        return suffdict[syl][0]
    # else try without the first letter
    elif syl[1:] in suffdict:
        return suffdict[syl[1:]][0]
    # else try without the first 2 letters
    elif syl[2:] in suffdict:
        return suffdict[syl[2:]][0]
    # else return None, which the calling function should check for
    else:
        return None

syl_re = re.compile("([bcdfghjklmnpqrstvwxz]{1,2}[aeiouy]+[bcdfghjklmnpqrstvwxz]*(e|ed)?('[a-z]{1,2})?)(?![a-zA-Z]+)")
#Returns phonemes of the word in an array. If word isn't in CMUDict, it tries to match the last few letters
#to a word that is. In that case it just returns the last phoneme. If it can't match at all, returns False.
def phonemes(word):
    word = re.split(r"[-_ ,]", word)[-1]#Make it use the last word if multiple words were passed in. TODO: probably a faster way to do this
    word = word.lower()
    # If in cmudict, just use cmudict
    if word in d:
        return d[word][0]

    # If not, try to use my cmu-based last syllable dictionary

    # if we cannot detect the last syllable, give up
    
    if not syl_re.search(word):
        return False
    last_syl = syl_re.search(word).group()

    # now try the last syllable against cmusuffdict
    p = try_syllable(last_syl)
    if p:
        return p
    # else try without the last 2 letters, if it ends in 's
    elif last_syl[-2:] == "'s":
        p = try_syllable(last_syl[:-2])
        if p:
            return p.append('Z')
        else:
            return False
    # else try without the last letter, if it ends in s
    elif last_syl[-1] == "s":
        p = try_syllable(last_syl[:-1])
        if p:
            return p.append('Z')
        else:
            return False
    else:  # If not in cmudict or my cmusuffdict
        return False

vowel_re = re.compile("[aeiouy]+")
falseSyllables_re = re.compile("(?<=\w)(ion|ious|(?<!t)ed|es|[^lr]e)(?![a-z']+)")

def approx_nsyl(word): #pass in a lowercase single word, and this will estimate its syllables, basically by counting verb islands.
    if word == "n't":#parser puts this by itself
        return 1
    digraphs = {"ai", "au", "ay", "ea", "ee", "ei", "ey", "oa", "oe", "oi", "oo", "ou", "oy", "ua", "ue", "ui"}#these are groups of vowels that only make one syllable. Other groups of vowels will be assumed to be two syllables.
    # Ambiguous, currently split: ie, io
    # Ambiguous, currently kept together: ui
    count = 0
    array = vowel_re.findall(word)#get a list of vowel islands
    for v in array:
        count+=1
        if len(v) > 1 and v not in digraphs:#group of vowels that makes more than one sound; assume it makes two sounds.
            count += 1
    if len(word)>=2 and word[-2:] == "es":
        count -=1
    
#    if falseSyllables_re.search(word): #These seem like they will be wrong about as often as right, so I'll skip them
#        count -= 1
#    if re.search("'ve|n't", word.lower()):
#        count += 1
    return count

#returns number of syllables in the word.
def nsyl(word):
    # return the min syllable count in the case of multiple pronunciations
    if not word.lower() in d:
        return approx_nsyl(word.lower())
    phonemes = d[word.lower()][0]
    return min([len([y for y in x if isVowel(y)]) for x in phonemes])
    # For example: d["concatenate".lower()] == [['K', 'AH0', 'N', 'K', 'AE1', 'T', 'AH0', 'N', 'EY2', 'T']]
    # Oh, and those numbers are actually stress/inflection (0: no stress, 1: primary stress, 2: secondary stress)
    # This grabs each item where the last character is a digit (how cmudict represents vowel sounds), and counts them

#returns number of syllables in the input text. Correctly handles text with spaces (e.g. if a wordnet synonym is really 2 words)
def nsylHandlingSpaces(text):
    nsyl = 0
    text = text.lower()
    for word in text.split():
        # return the min syllable count in the case of multiple pronunciations
        if not word in d:
            nsyl += approx_nsyl(word)
        else:
            phonemes = d[word][0]
            for phoneme in phonemes:
                if isVowel(phoneme):
                    nsyl += 1
    return nsyl

# Ignores stress for now, while I'm not taking meter into account
# Returns True if the last syllable is a match
def rhyme_from_phonemes(list1, list2):
    i = -1
    while i >= 0 - len(list1):#loop back until we reach a vowel
        if list1[i][-1].isdigit():#because vowels end in digits
            if i >= 0 - len(list2) and list1[i][:-1] == list2[i][:-1] and (i == -1 or list1[i + 1:] == list2[i + 1:]):
                return True
            else:
                return False
        i -= 1

#true if a cmudict phoneme is a vowel
def isVowel(phoneme):
    return phoneme[-1].isdigit()

def vowelsMatch(phonemeA, phonemeB):
    #return phonemeA[:-1] == phonemeB[:-1] #to ignore stress
    return phonemeA == phonemeB #to require stress to match

#Returns a float score, where numbers >= 1 are rhymes, and higher numbers are better rhymes.
#TODO: Change this to capture that close vowel sounds are close. Right now it rejects unless they're exact.
#TODO: Don't count a similar "er" ending as a rhyme. E.g. fighter and roadster don't rhyme
def rhyme_score_from_phonemes(list1, list2):
    MIN_RHYME_SCORE=1 #values >= this indicate rhymes 
    CONSONANT_VALUE=.3 #bonus for internal consonants
    VOWEL_VALUE = .9
    VOWEL_STRESS_MATCH_BONUS = .29 #bonus for the vowel stresses matching. Must be less than CONSONANT_VALUE
    LOOSE_INTERNAL_CONSONANT_MATCHING=True #if true, consonants to the left of the first vowel don't have to match exactly. (It still helps the score if they do) 
    index1 = -1#we'll index from the end of the list
    score=0

    #rightmost syllable is a special case; it must match perfectly
    while index1 >= -len(list1) :#loop back until we reach a vowel
        if isVowel(list1[index1]):
            if index1 >= -len(list2) and vowelsMatch(list1[index1], list2[index1]): #list1's ending vowel matches a vowel the same dist from the end of list 2
                stressMatchBonus = 0
                if list1[index1][-1] == list2[index1][-1]:
                    stressMatchBonus = VOWEL_STRESS_MATCH_BONUS
                if index1 == -1 :#we're still on the very rightmost phoneme
                    score = MIN_RHYME_SCORE+stressMatchBonus-CONSONANT_VALUE#a matching vowel at the end of the words will require an additional consonant to be considered a rhyme. Otherwise it says toyota rhymes with brahma, for example.
                    break
                elif list1[index1+1:] == list2[index1+1:]:#all the consonant sounds to the right match
                    score = VOWEL_VALUE+stressMatchBonus #+ CONSONANT_VALUE*(-index1-1)
                    break
            return False
        index1 -= 1
    
    index1 -= 1
    index2 = index1
    if index1 < -len(list1) or index2 < -len(list2) :
        return score
    #first syllable was a rhyme, now look at internal syllables to boost its score.  
    while True:
        #while we have matching consonants add up their scores, and move both indices back in step
        while not isVowel(list1[index1]) and not isVowel(list2[index2]):
            if list1[index1] == list2[index2]:#consonants match
                score += CONSONANT_VALUE
                index1-=1
                index2-=1
                if index1 < -len(list1) or index2 < -len(list2) :
                    return score
            else:#consonants don't match
                if LOOSE_INTERNAL_CONSONANT_MATCHING:#it's ok that they don't match
                    break
                else:#it's not ok that they don't match
                    return score
        
        if LOOSE_INTERNAL_CONSONANT_MATCHING:#skip any further consonants to get to a vowel
            while not isVowel(list1[index1]):
                index1-=1
                if index1 < -len(list1):
                    return score
            while not isVowel(list2[index2]):
                index2-=1
                if index2 < -len(list2):
                    return score
                
        #if we're here without returning, we've found the next two vowels
        if not vowelsMatch(list1[index1], list2[index2]):#the vowels don't match
            return score
        #else, the vowels match
        score += VOWEL_VALUE
        if list1[index1][-1] == list2[index2][-1]:#the stresses also match
            score += VOWEL_STRESS_MATCH_BONUS
        index1-=1
        index2-=1
        if index1 < -len(list1) or index2 < -len(list2) :
            return score


def rhyme(word1, word2):
    list1 = min(d[word1.lower()], key=len)
    list2 = min(d[word2.lower()], key=len)
    return rhyme_from_phonemes(list1, list2)


def tokenize_text(text):
    text = re.sub("[^a-zA-Z\s'-]", '', text)
    text = re.sub("'(?![a-z]{1,2})", '', text)
    tokens = re.split("\s+|-", text)
    # remove empty tokens
    tokens = filter(None, tokens)
    return tokens


def tokenize(file_path):
    with open(file_path) as f:
        data = f.read().strip()
    return tokenize_text(data)


def tokenize_from_url(url):
    data = urllib2.urlopen(url).read().strip()
    return tokenize_text(data)

# Thinking about meter:
# In "there once" [was a man from Nantucket], I'd want to see that "there" is unstressed, and "once" is stressed
# But cmudict sees the single vowel in each of them as 1 (primary stress), because it looks at each word in isolation
# Maybe for now just assume than monosyllabic words are flexible, and use cmudict for stress on polysyllabic words?
